<?php

namespace Drupal\splio_utils\Plugin\rest\resource;

use Drupal\rest\Plugin\ResourceBase;
use Drupal\rest\ResourceResponse;
use Symfony\Component\HttpFoundation\Response;

/**
 * Class SplioBlacklistResource.
 *
 * Provides a REST service for the SplioBlacklistManager service.
 *
 * @RestResource(
 *   id = "splio_blacklist_manager",
 *   label = @Translation("Splio blacklist manager"),
 *   uri_paths = {
 *     "canonical" = "/splio/blacklist/{email}",
 *     "create" = "/splio/blacklist",
 *   }
 * )
 */
class SplioBlacklistResource extends ResourceBase {

  /**
   * Returns whether an email address is blacklisted in Splio.
   *
   * @param string $email
   *   The email address to be checked.
   *
   * @return \Drupal\rest\ResourceResponse
   *   - status_code
   *   - message:
   *    {
   *      "blacklisted":
   *        "true" if email address is blacklisted; "false" otherwise.
   *    }
   */
  public function get(string $email): ResourceResponse {

    $response = new ResourceResponse();
    /** @var \Drupal\splio_utils\Services\SplioBlacklistManager $blacklistManager */
    $blacklistManager = \Drupal::service('splio_utils.splio_blacklist_manager');
    $isBlacklisted = $blacklistManager->isEmailBlacklisted($email);

    if ($isBlacklisted) {
      $response->setStatusCode(Response::HTTP_OK);
    }
    else {
      $response->setStatusCode(Response::HTTP_NOT_FOUND);
    }

    $response->setContent(
      json_encode(["blacklisted" => $isBlacklisted])
    );

    $response->addCacheableDependency(
      [
        '#cache' => [
          'max-age' => 0,
        ],
      ]
    );

    return $response;
  }

  /**
   * Adds various email addresses to the Splio Blacklist.
   *
   * @param array $data
   *   The email addresses to be blacklisted.
   *
   * @return \Drupal\rest\ResourceResponse
   *   Response with results:
   *   [
   *      'successes': array of emails blacklisted,
   *      'failures': array of emails not blacklisted,
   *   ]
   */
  public function post(array $data): ResourceResponse {

    $resourceResponse = new ResourceResponse();
    /** @var \Drupal\splio_utils\Services\SplioBlacklistManager $blacklistManager */
    $blacklistManager = \Drupal::service('splio_utils.splio_blacklist_manager');
    $response = $blacklistManager->addEmailsToBlacklist($data);

    if (empty($response)) {
      $resourceResponse->setStatusCode(Response::HTTP_INTERNAL_SERVER_ERROR);
    }
    elseif (!empty($response[0]['status_code'])) {
      $resourceResponse->setStatusCode($response[0]['status_code']);
    }

    if (!empty($response[0]['message'])) {
      $resourceResponse->setContent($response[0]['message']);
    }

    return $resourceResponse;
  }

}
