<?php

namespace Drupal\splio_utils\Services;

use Drupal\splio\Services\SplioConnectorInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Class SplioBlacklistManager.
 *
 * Manages the blacklist of the Splio platform. Allows to check if an email
 * address is blacklisted in Splio and add any email address to the Blacklist.
 *
 * Due to Splio API limitations, an email address cannot be removed form the
 * blacklist through this module.
 */
class SplioBlacklistManager {

  /**
   * Connector to Splio services.
   *
   * @var \Drupal\splio\services\SplioConnectorInterface
   */
  private SplioConnectorInterface $splioConnector;

  /**
   * SplioBlacklistManager constructor.
   *
   * @param \Drupal\splio\services\SplioConnectorInterface $splioConnector
   *   Connector to Splio services.
   */
  public function __construct(SplioConnectorInterface $splioConnector) {

    $this->splioConnector = $splioConnector;
  }

  /**
   * SplioBlacklistManager create method.
   *
   * @param \Symfony\Component\DependencyInjection\ContainerInterface $container
   *   Main container.
   *
   * @return \Drupal\splio_utils\Services\SplioBlacklistManager
   *   Returns an instance of SplioBlacklistManager with the injected services.
   */
  public static function create(
    ContainerInterface $container
  ): SplioBlacklistManager {

    return new static(
      $container->get('splio.splio_connector')
    );
  }

  /**
   * Checks whether an email address is Blacklisted in Splio or not.
   *
   * @param string $email
   *   The email address to check.
   *
   * @return bool
   *   Returns TRUE in case the received email address is Blacklisted in Splio.
   *   It will return FALSE in any other case.
   */
  public function isEmailBlacklisted(string $email): bool {

    return $this->splioConnector->isEmailBlacklisted($email);
  }

  /**
   * Adds an email address to the user's configured universe Blacklist in Splio.
   *
   * @param string[] $emails
   *   The email address to be blacklisted.
   *
   * @return array
   *   Returns TRUE in case the received email address was successfully
   *   blacklisted. Return FALSE in any other case.
   */
  public function addEmailsToBlacklist(array $emails): array {

    return $this->splioConnector->addEmailsToBlacklist($emails);
  }

}
