<?php

namespace Drupal\splio_utils\Services;

use Drupal\splio\Services\SplioTriggerConnector;
use GuzzleHttp\Exception\GuzzleException;
use Psr\Log\LoggerInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Class SplioTriggerManager.
 *
 * Manages the emailing triggers of the Splio platform. Allows the user to
 * trigger the delivery for a certain message to a defined set of contacts
 * received by parameter.
 */
class SplioTriggerManager {

  /**
   * Logger.
   *
   * @var \Psr\Log\LoggerInterface
   */
  private LoggerInterface $logger;

  /**
   * Connector to Splio trigger service.
   *
   * @var \Drupal\splio\Services\SplioTriggerConnector
   */
  private SplioTriggerConnector $splioTriggerConnector;

  /**
   * SplioTriggerManager constructor.
   *
   * @param \Psr\Log\LoggerInterface $logger
   *   The logger. Loads the 'splio' channel.
   * @param \Drupal\splio\Services\SplioTriggerConnector $splioTriggerConnector
   *   Connector to Splio.
   */
  public function __construct(
    LoggerInterface $logger,
    SplioTriggerConnector $splioTriggerConnector
  ) {

    $this->logger = $logger;
    $this->splioTriggerConnector = $splioTriggerConnector;
  }

  /**
   * SplioTriggerManager create method.
   *
   * @param \Symfony\Component\DependencyInjection\ContainerInterface $container
   *   Main container.
   *
   * @return \Drupal\splio_utils\Services\SplioTriggerManager
   *   Returns an instance of SplioTriggerManager with the injected services.
   */
  public static function create(ContainerInterface $container) {

    return new static(
      $container->get('logger.channel.splio'),
      $container->get('splio.splio_trigger_connector')
    );
  }

  /**
   * Sends a trigger to Splio so the platform sends a message to the recipients.
   *
   * Receives a message_id and a set of users (a string of recipients with its
   * own properties formatted as JSON). Sends a request to Splio to trigger the
   * message delivery.
   *
   * @param array $data
   *   The trigger data to be fired in Splio.
   *   [
   *      message => 'xxxxx',
   *      opcode => [
   *        type => 'xxxx',
   *        success_message => 'xxx',
   *        failure_message => 'xxx',
   *      ],
   *   ].
   *
   * @return array
   *   [
   *     code => HTTP response status code,
   *     error => has value if an error occurred
   *   ].
   */
  public function triggerMessage(array $data): array {

    return $this->splioTriggerConnector->triggerMessage($data);
  }

}
