<?php

namespace Drupal\splio\Services;

use Drupal\Core\Config\ConfigFactory;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Queue\SuspendQueueException;
use Drupal\splio\Entity\SplioEntity;
use Drupal\splio\Exception\EntitiesNotConfiguredYetException;
use Drupal\splio\Exception\EntityTypeNotFoundInSplioException;
use Psr\Log\LoggerInterface;

/**
 * Splio entity handler.
 */
class SplioEntityHandler implements SplioEntityHandlerInterface {

  const PREFIX_ORDER_LINES = 'order_lines_';

  /**
   * Entity type manager service.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  private EntityTypeManagerInterface $entityTypeManager;

  /**
   * Entity value handler.
   *
   * @var \Drupal\splio\Services\EntityValueHandlerInterface
   */
  private EntityValueHandlerInterface $entityValueHandler;

  /**
   * Logger service.
   *
   * @var \Psr\Log\LoggerInterface
   */
  private LoggerInterface $logger;

  /**
   * Config factory service.
   *
   * @var \Drupal\Core\Config\ConfigFactory
   */
  private ConfigFactory $config;

  public function __construct(
    EntityTypeManagerInterface $entityTypeManager,
    EntityValueHandlerInterface $entityValueHandler,
    LoggerInterface $logger,
    ConfigFactory $config
  ) {

    $this->entityTypeManager = $entityTypeManager;
    $this->entityValueHandler = $entityValueHandler;
    $this->logger = $logger;
    $this->config = $config;
  }

  /**
   * Load Splio entities from local storage.
   *
   * @return array|mixed|null
   *   Splio type entities.
   *
   * @throws \Drupal\Core\Queue\SuspendQueueException
   */
  public function getLocalSplioEntitiesConfig() {

    $entitiesConfig = $this->config
      ->get('splio.entity.config')
      ->get('splio_entities');

    if (empty($entitiesConfig)) {
      $message = 'You have not configured the relation of your entities with Splio ones yet.';

      $this->logger->error($message);

      throw new SuspendQueueException($message);
    }

    return $entitiesConfig;
  }

  /**
   * Gets local Splio entities fields.
   *
   * @param array $propertiesFilter
   *   ['splio_entity' => 'SOME_SPLIO_ENTITY_TYPE'].
   *
   * @return \Drupal\Core\Entity\EntityInterface[]
   *   List of entity fields.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  public function getLocalSplioEntitiesFields(
    array $propertiesFilter
  ): array {

    return $this->entityTypeManager
      ->getStorage('splio_field')
      ->loadByProperties($propertiesFilter);
  }

  /**
   * Determines if an entity is configured by the user as a Splio entity.
   *
   * Helper function that determines if a received entity belongs to the Splio
   * configured entities. Returns the Splio entity type if true, throws an
   * exception otherwise.
   *
   * @param \Drupal\Core\Entity\EntityInterface|null $entity
   *   The entity that may belong to Splio.
   *
   * @return string|null
   *   Returns a string with the SplioEntity type in case this entity is
   *   configured as a Splio entity. Returns FALSE otherwise.
   *
   * @throws \Drupal\splio\Exception\EntityTypeNotFoundInSplioException
   * @throws \Drupal\Core\Queue\SuspendQueueException
   */
  public function getSplioEntityTypeFromLocalEntity(
    ?EntityInterface $entity
  ): ?string {
    $splioEntities = $this->getLocalSplioEntitiesConfig();

    foreach ($splioEntities as $type => $definition) {
      if ($entity->getEntityType()->id() === $definition['local_entity']) {
        return $type;
      }
    }

    throw new EntityTypeNotFoundInSplioException($entity->getEntityType()->id());
  }

  /**
   * Gets the proper entity type.
   *
   * @param string $entityType
   *   Given entity type.
   *
   * @return string
   *   "receipts" if given entity type is "order_lines".
   *   The given entity type itself, otherwise.
   */
  public function getProperEntityType(string $entityType): string {

    if (SplioEntity::TYPE_ORDER_LINES === $entityType) {
      $entityType = SplioEntity::TYPE_RECEIPTS;
    }

    return $entityType;
  }

  /**
   * Gets the proper entity.
   *
   * @param string $entityType
   *   Entity type.
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   Entity.
   *
   * @return \Drupal\Core\Entity\EntityInterface|null
   *   An order entity object, if given entity type is type of "order_lines".
   *   The given entity itself, otherwise.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  public function getProperEntity(
    string $entityType,
    EntityInterface $entity
  ): ?EntityInterface {

    if (SplioEntity::TYPE_ORDER_LINES === $entityType) {
      $orderEntity = $this->getOrderFromOrderLine($entity);

      if (!empty($orderEntity)) {
        $original = $entity;
        $entity = $orderEntity;
        $entity->original = $original;
      }
      else {
        $this->logger->warning(
          'Could not retrieve an order for the received %entity order_lines entity.',
          [
            '%entity' => $entity->getEntityTypeId(),
          ]
        );

        return NULL;
      }
    }

    return $entity;
  }

  /**
   * Gets the value of an entity key field.
   *
   * @param string $entityType
   *   Entity type.
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   Entity.
   *
   * @return string
   *   Key field value.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  public function getEntityKeyFieldValue(
    string $entityType,
    EntityInterface $entity
  ): string {

    $keyField = $this->getEntityTypeKeyField($entityType);

    return $entity->get($keyField)->value;
  }

  /**
   * Gets and entity type key field.
   *
   * @param string $entityType
   *   Entity type.
   *
   * @return string
   *   Key field name.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  public function getEntityTypeKeyField(string $entityType): string {

    $fields = $this->getLocalSplioEntitiesFields(
      ['splio_entity' => $entityType]
    );

    /** @var \Drupal\splio\Entity\SplioField $field */
    foreach ($fields as $field) {
      if ($field->isKeyField()) {
        return $field->getDrupalField();
      }
    }

    return '';
  }

  /**
   * Gets the Drupal field name from an entity type field.
   *
   * @param string $entityType
   *   Entity type.
   * @param string $fieldName
   *   Field name.
   *
   * @return string
   *   Drupal field name.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  public function getDrupalFieldFromSplioFieldEntity(
    string $entityType,
    string $fieldName
  ): string {

    $drupalField = '';
    $fields = $this->getLocalSplioEntitiesFields(
      ['splio_entity' => $entityType]
    );

    /** @var \Drupal\splio\Entity\SplioFieldInterface $field */
    foreach ($fields as $field) {
      if ($field->getDrupalField() === $fieldName) {
        $drupalField = $field->getDrupalField();

        break;
      }
    }

    return $drupalField;
  }

  /**
   * Receives an order_line entity and returns the receipt that it belongs to.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   The received order_line entity.
   *
   * @return \Drupal\Core\Entity\EntityInterface|null
   *   Returns a receipt (order) entity.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  private function getOrderFromOrderLine(
    EntityInterface $entity
  ): ?EntityInterface {

    // Retrieve the Drupal field that contains the order id for this
    // order line item.
    $splioEntityFields = $this->getLocalSplioEntitiesFields(
      ['splio_entity' => SplioEntity::TYPE_ORDER_LINES]
    );
    $orderIdDrupalField = $splioEntityFields[self::PREFIX_ORDER_LINES . 'id_order']
      ->getDrupalField();

    // Next, obtain the value of the Order id,
    // which this order line belongs to.
    if (!empty($orderIdDrupalField)) {
      $orderIdValue = $this->entityValueHandler->getFieldValue(
        $orderIdDrupalField,
        $entity
      );
    }
    else {
      $this->logger
        ->error("Error trying to fetch Receipts key field. Check that your Receipts key field is configured properly.");
      return NULL;
    }

    $orderIdValue = is_array($orderIdValue)
      ? $orderIdValue[array_keys($orderIdValue)[0]] : $orderIdValue;

    $orderEntityType = '';

    if (!empty($orderIdValue)) {
      try {
        $orderKeyFieldData = $this->getLocalSplioEntitiesFields(
          [
            'splio_entity' => SplioEntity::TYPE_RECEIPTS,
            'is_key_field' => '1',
          ]
        );

        if (1 !== \count($orderKeyFieldData)) {
          throw new \Exception();
        }

        $entities = $this->getLocalSplioEntitiesConfig();
        $orderEntityType = $entities[SplioEntity::TYPE_RECEIPTS]['local_entity'];

        /** @var \Drupal\splio\Entity\SplioField $orderKeyField */
        $orderKeyField = end($orderKeyFieldData);
        $orderKeyDrupalField = $orderKeyField->getDrupalField();
      }
      catch (\Exception $exception) {
        $this->logger
          ->error("Error trying to fetch Receipts key field. Check that your Receipts key field, and the 'id_order' field from Order Lines, are configured properly.");
      }

      // Finally, load the proper $orderEntityType by passing the
      // $orderIdValue as the $orderKeyField.
      if (!empty($orderKeyDrupalField)) {
        $entities = $this->entityTypeManager
          ->getStorage($orderEntityType)
          ->loadByProperties([$orderKeyDrupalField => $orderIdValue]);

        return $entities[$orderIdValue] ?? NULL;
      }
    }

    return NULL;
  }

}
