<?php

namespace Drupal\splio\Services;

use Drupal\Component\EventDispatcher\ContainerAwareEventDispatcher;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Queue\QueueFactory;
use Drupal\splio\Entity\SplioEntity;
use Drupal\splio\Event\SplioQueueEvent;
use Drupal\splio\Exception\EntityTypeNotFoundInSplioException;
use Psr\Log\LoggerInterface;

/**
 * Handles the Splio items in Drupal queue.
 */
class SplioQueueHandler implements SplioQueueHandlerInterface {

  /**
   * Splio entity handler.
   *
   * @var \Drupal\splio\Services\SplioEntityHandlerInterface
   */
  private SplioEntityHandlerInterface $splioEntityHandler;

  /**
   * Queue factory.
   *
   * @var \Drupal\Core\Queue\QueueFactory
   */
  private QueueFactory $queueFactory;

  /**
   * Event dispatcher.
   *
   * @var \Drupal\Component\EventDispatcher\ContainerAwareEventDispatcher
   */
  private ContainerAwareEventDispatcher $eventDispatcher;

  /**
   * Logger.
   *
   * @var \Psr\Log\LoggerInterface
   */
  private LoggerInterface $logger;

  /**
   * Entity helper.
   *
   * @var \Drupal\splio\Services\EntityHelperInterface
   */
  private EntityHelperInterface $entityHelper;

  const SPLIO_QUEUE_ID = 'cron_splio_sync';

  const VALID_ACTIONS = [
    'create' => 'create',
    'update' => 'update',
    'delete' => 'delete',
    'dequeue' => 'dequeue',
  ];

  public function __construct(
    SplioEntityHandlerInterface $splioEntityHandler,
    QueueFactory $queueFactory,
    ContainerAwareEventDispatcher $eventDispatcher,
    LoggerInterface $logger,
    EntityHelperInterface $entityHelper
  ) {

    $this->splioEntityHandler = $splioEntityHandler;
    $this->queueFactory = $queueFactory;
    $this->eventDispatcher = $eventDispatcher;
    $this->logger = $logger;
    $this->entityHelper = $entityHelper;
  }

  /**
   * Adds the received splio entity to a queue to be processed later.
   *
   * Receives a splio entity and the CRUD action to perform with it.
   * The entity will be added to a queue where it will be processed on the
   * next cron run.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   The entity which will be added to the queue.
   * @param string $action
   *   CRUD action that will be performed when the received entity is processed.
   *   Accepts the following parameters: create, update, delete and dequeue.
   */
  public function addEntityToQueue(EntityInterface $entity, $action) {

    try {
      $entityType = $this->splioEntityHandler
        ->getSplioEntityTypeFromLocalEntity($entity);
    }
    catch (EntityTypeNotFoundInSplioException $exception) {
      return;
    }

    $item = $this->generateQueueItem($entityType, $entity, $action);

    // Manage the event to be dispatched.
    $queueEvent = new SplioQueueEvent($item);
    $this->eventDispatcher->dispatch(
      $queueEvent,
      SplioQueueEvent::SPLIO_ENQUEUE
    );

    // In case someone captured the event and made changes in the item,
    // update the item before inserting it into the queue.
    $item = $queueEvent->getSplioQueueItem();

    // Finally, perform a last check to ensure the set action is valid.
    if ('dequeue' === $item['action']) {
      return;
    }

    $queue = $this->queueFactory->get(self::SPLIO_QUEUE_ID);

    if (in_array($item['action'], self::VALID_ACTIONS)) {
      $queue->createItem($item);

      return;
    }

    $this->logger->error(
      "The %type[%id] entity will not be queued. Action type received: %action. Only 'create', 'update' and 'delete' actions are queued.",
      [
        '%action' => $item['action'],
        '%type' => $item['splioEntityType'],
        '%id' => $item['id'],
      ]
    );
  }

  /**
   * Deletes an item from queue.
   *
   * @param array $item
   *   Item to delete.
   */
  public function deleteItem(array $item) {

    $queue = $this->queueFactory->get(self::SPLIO_QUEUE_ID);
    $queue->deleteItem($item);
  }

  /**
   * Generates the item that will be queued later.
   *
   * @return array
   *   Queue item.
   */
  private function generateQueueItem(
    string $entityType,
    EntityInterface $entity,
    string $action
  ) {

    $item = [];
    $entity = $this->splioEntityHandler->getProperEntity($entityType, $entity);

    if (SplioEntity::TYPE_ORDER_LINES === $entityType) {
      $entityType = SplioEntity::TYPE_RECEIPTS;
      $item += ['originalSplioEntityType' => SplioEntity::TYPE_ORDER_LINES];
    }

    // In case there is an original entity (update), add it to the queue item.
    // In case there is a delete action, add the deleted item as original.
    $original = $this->entityHelper->getOriginal($entity);

    if ('delete' === $action) {
      $original = $entity;
    }

    $item += [
      'id' => $entity->id(),
      'original' => $original,
      'lang' => $entity->language()->getId(),
      'splioEntityType' => $entityType,
      'action' => $action,
    ];

    return $item;
  }

}
